#### Load libraries ####
rm(list=ls())
library(MASS) # mvrnorm()
library(clusterGeneration) # genPositiveDefMat()
library(statmod)
library(rlist)
library(dplyr)
library(purrr) # map
library(survival) # coxph


#### Set directories and load files ####
jobname <- "LMA_simdata_GenerateData"

maindir <- "/nas/longleaf/home/jitong/thesis/latent_model_all"
codes_dir <- file.path(maindir,"codes")
wsdir <- file.path(maindir,"workspaces",jobname)

source(file.path(codes_dir,"functions_needed_GHinput_ManyBiomarkers.R"))

#### Settings of variables and parameters ####
N <- 5000 # number of patients
m <- 2 # number of baseline variables
p <- 6 # number of temporal processes(k)

## a (m+1)*p matrix, each column is beta for each temporal process(at time t)
beta_kt <- function(t){matrix(c(-0.44-t/8, 0.6+t^(1/2)/3, -0.5+t^(1/3)/2,
                                -0.93+t/9, -0.53-t^(1/2)/2, 0.4+t^(1/3)/2, 
                                0.35+t/10, -2+t^(1/2), 1.9-t^(1/3), 
                                -1.36+t/10, sin(0.76+t), cos(-0.3+t),
                                cos(-0.25+t), 0.37+t/10, sin(-0.68+t),
                                0.91+(t-6)^3/216, t/10, 1.23+(t-6)^2/36
), ncol=p, byrow=F)}

## a p*p matrix, the covariance matrix for epsilons(random effect) at each fixed time point
set.seed(666)
Sigma1_t <- genPositiveDefMat(dim=p, covMethod="unifcorrmat", rangeVar=c(1,1), lambdaLow=1, ratioLambda=10)[[2]]
sigma_sq_all <- diag(Sigma1_t)

## Canonical links, their inverse functions, and their derivatives for each response(latent temporal process) 
par0.glm.func <- list(gaussian(), poisson(), binomial(),
                      gaussian(), binomial(), binomial()
)

linkfunc.inv <- list(identity, exp, function(z){exp(z)/(1+exp(z))},
                     identity, function(z){exp(z)/(1+exp(z))}, 
                     function(z){exp(z)/(1+exp(z))}
)

linkfunc.inv.deriv <- list(function(x){rep(1,length(x))}, exp, function(z){exp(z)/(1+exp(z))^2},
                           function(x){rep(1,length(x))}, function(z){exp(z)/(1+exp(z))^2}, 
                           function(z){exp(z)/(1+exp(z))^2}
)

## Time of interest
M <- 12
t.grids <- 12*(0:M)/M

#### Generate measured timepoints ####
## a m*p matrix, each column contains the gamma parameters for each temporal response
gamma_k <- matrix(rep(c(0.5, 0.25), times=p), nrow = m, byrow = FALSE)

## a list of length p, each element is a function of t for each temporal response
integrals.lambda_kt <- as.list(rep(0.5, p))

# Create characteristics of patients, Z_i, a N*(m+1) matrix, each row for each patient
rseed <- 1
set.seed(rseed)
T_i <- rep(12, N) # used in time.measured()
bv <- data.frame(i=1:N,
                int=rep(1, N),
                age=runif(N, min=-1, max=1),
                sex=rbinom(n=N ,size=1, prob=0.5)-0.5)
Z_i <- bv %>% select(-i) %>% as.matrix()
print(head(Z_i))

## create measured timepoints for each patient, each process
# a list of length N, each element is a list of length p, each element of the sublist is the measured time points for the corresponding patient and process
time.measured.all <- vector("list",N)
names(time.measured.all) <- paste0("i=",seq(1,N))

for (i in 1:N){
  time.measured.all[[i]] <- vector("list", p)
  names(time.measured.all[[i]]) <- paste0("k=",seq(1,p))

  for (k in 1:4){
    time.measured.all[[i]][[k]] <- time.measured(i, k, integrals.lambda_kt[[k]])
  }
  time.measured.all[[i]][[5]] <- time.measured.all[[i]][[3]] + c(0, rnorm(length(time.measured.all[[i]][[3]])-1, mean=0, sd=0.05))
  time.measured.all[[i]][[5]] <- sort(time.measured.all[[i]][[5]][time.measured.all[[i]][[5]]>=0 & time.measured.all[[i]][[5]]<=12])

  time.measured.all[[i]][[6]] <- time.measured.all[[i]][[3]] + c(0, rnorm(length(time.measured.all[[i]][[3]])-1, mean=0, sd=0.05))
  time.measured.all[[i]][[6]] <- sort(time.measured.all[[i]][[6]][time.measured.all[[i]][[6]]>=0 & time.measured.all[[i]][[6]]<=12])

  for (k in 1:p){
    names(time.measured.all[[i]][[k]]) <- paste0("j=",seq(1,length(time.measured.all[[i]][[k]])))
  }

  if (i%%1000==0){print(paste(i, "is done"))}
}

## check the distribution of time points(value and count)
for (sss in 1:p){
  print(summary(unlist(lapply(time.measured.all, function(z){z[[sss]]})))) # value
  print(summary(sapply(lapply(time.measured.all, function(z){z[[sss]]}), length))) # count
}

#### Generate responses by generating epsilon first, then mu ####
b <- 1

## epsilon, random effect
# a list of length N, each element is should be a p*Ni matrix, each column is the value of epsilon for each process at a specific timepoint
epsilon.all <- lapply(
  lapply(
    lapply(time.measured.all,
           function(t.list){
             t_ip <- unlist(t.list)
             timepoints.cov.coef(t_ip, b) # currently might exist duplicated timepoints
           }),
    function(Sigma2_i){
      Sigma2_i %x% Sigma1_t  # A %x% B = (a_ij*B)
    }),
  function(Omega){
    matrix(MASS::mvrnorm(n=1, mu=rep(0,nrow(Omega)), Sigma=Omega), nrow=p, byrow=F)
  }
) # could delete Sigma2.all and Omega_t_inverse.all to save space

## mu, mean of the exponential family
# a list of length N, each element should be a p*Ni matrix, each column is values of mu of all processes at a specific timepoint
mu_ik.all <- vector("list",N)
names(mu_ik.all) <- names(time.measured.all)
for (i in 1:N){
  mu_ik.all[[i]] <- sapply(unlist(time.measured.all[[i]]), function(t_ikj){
    t(beta_kt(t_ikj)) %*% (Z_i[i,]) # a p*Ni matrix, each column is t(beta)*Z of all processes at a specific timepoint
  }) + epsilon.all[[i]]
}

## response variable
# a list of length N, each element should be a p*Ni matrix, each column is responses of all processes at a specific timepoint
norm.var <- 0.5 # phi(variance) for normal distributions
X_ikj.all <- lapply(mu_ik.all, function(mu_ik){
  out <- rbind(matrix(mvrnorm(1, mu_ik[1,], norm.var*diag(ncol(mu_ik))), nrow = 1),
               matrix(rpois(ncol(mu_ik), exp(mu_ik[2,])), nrow=1),
               matrix(rbinom(n=ncol(mu_ik), size=1, prob=exp(mu_ik[3,])/(1+exp(mu_ik[3,]))), nrow=1),
               matrix(mvrnorm(1, mu_ik[4,], norm.var*diag(ncol(mu_ik))), nrow = 1),
               matrix(rbinom(n=ncol(mu_ik), size=1, prob=exp(mu_ik[5,])/(1+exp(mu_ik[5,]))), nrow=1),
               matrix(rbinom(n=ncol(mu_ik), size=1, prob=exp(mu_ik[6,])/(1+exp(mu_ik[6,]))), nrow=1)
  )
  colnames(out) <- colnames(mu_ik) # xxx is a matrix, having colnames, but xxx[1,] is a vector which only has names
  return(out) # a p*Ni matrix, [k,j] element is the response for k-th temporal process and j-th timepoint
})

### Combine measured time points and biomarkers, convert lists to a matrix
time_num <- lapply(time.measured.all, function(t_i){sapply(t_i, length)})
time_val <- lapply(time.measured.all, unlist)

combine_time_Xikj_all <- function(i, t_num, t_val, X_mat, mu_mat, epsilon_mat){
  out <- matrix(0, nrow=ncol(X_mat), ncol=7)
  out[,1] <- i
  out[,2] <- rep(1:p, t_num)
  out[,3] <- unlist(sapply(t_num, function(n){1:n}))
  out[,4] <- t_val
  out[,5] <- t(X_mat)[cbind(1:ncol(X_mat), out[,2])]
  out[,6] <- t(mu_mat)[cbind(1:ncol(mu_mat), out[,2])]
  out[,7] <- t(epsilon_mat)[cbind(1:ncol(epsilon_mat), out[,2])]
  return(out)
}

time_Xikj_all_list <- mapply(combine_time_Xikj_all, 1:N, time_num, time_val, X_ikj.all, mu_ik.all, epsilon.all) # arguments are lists, whose element shouldn't be a list, otherwise it won't generate the expected results
time_Xikj_all_mat <- do.call(rbind, time_Xikj_all_list)[,1:5]
colnames(time_Xikj_all_mat) <- c("i","k","j","time","value")
time_Xikj_all_df <- as.data.frame(time_Xikj_all_mat)

## Create data for fitting Coxph model (Andersen-Gill) and estimate gamma ####
df_coxph <- map_dfr(time.measured.all, function(t_i){
  map_dfr(t_i, function(t_ik){
    data.frame(start = t_ik,
               stop = c(t_ik[-1],12),
               event = c(rep(1,length(t_ik)-1),0))
  })
})

df_coxph$id <- time_Xikj_all_mat[,1]
df_coxph$marker <- time_Xikj_all_mat[,2]
df_coxph_AG <- cbind(df_coxph,Z_i[df_coxph$id,])

gamma.hat.all <- mapply(function(k){
  coxph(Surv(start, stop, event) ~ (age + sex) + cluster(id), 
        df_coxph_AG %>% filter(marker==k))$coefficient
}, k=1:p) # m*p matrix, each column for each temporal process

bv <- data.frame(i=1:N,Z_i)

colnames(time_Xikj_all_mat) <- c("i","k","j","time","value")
time_Xikj_all_df <- as.data.frame(time_Xikj_all_mat)

#### Sample correlation for each pair of biomarkers at each time point ####
cor_mat_lists <- vector("list",length=length(t.grids)) # need t.grids, time_time_Xikj_all_df
for (mmm in 1:length(t.grids)){
  y <- vector("list",length=p)
  ttt <- t.grids[mmm]
  for (i in 1:p){
    y[[i]] <- closest_Obs(time_Xikj_all_df, ttt, i)
  }
  cor_mat <- matrix(0, nrow=p, ncol=p)
  for (kkk in 1:(p-1)){
    for (lll in ((kkk+1):p)){  
      cor_mat[kkk,lll] <- cor(y[[kkk]],y[[lll]])
    }
  }
  cor_mat_lists[[mmm]] <- cor_mat
}
names(cor_mat_lists) <- paste0("t=",t.grids)
reduce(cor_mat_lists,`+`)/length(t.grids)

save(bv, time_Xikj_all_df, gamma.hat.all,cor_mat_lists,
     sigma_sq_all,t.grids,
     file=file.path(wsdir, paste0(jobname,"_seed",rseed,".RData")))