#### Load libraries ####
rm(list=ls())
library(statmod)
library(rlist)
library(dplyr)
library(purrr) # map
library(magrittr) # %<>%

#### Set directories ####
jobname <- "LMA_simdata_est"
cs <- paste0("case", 1)
rseed <- 1
rsd <- paste0("seed", 1)
casename <- paste(jobname,cs,rsd,sep="_")
wsname <- "LMA_simdata_GenerateData"

maindir <- "/nas/longleaf/home/jitong/thesis/latent_model_all"
paramdir <- file.path(maindir,"results",jobname,cs,rsd)
consoledir <- file.path(maindir,"Rconsole",jobname,cs,rsd)
iterdir <- file.path(maindir,"results",jobname,"temp")
codes_dir <- file.path(maindir,"codes")
wsdir <- file.path(maindir,"workspaces",wsname)

#### Load files ####
source(file.path(codes_dir,"functions_needed_simdata.R"))
load(file.path(wsdir,
               paste0(wsname,"_",rseed,".RData"))
     ) # bv, time_Xikj_all_mat, gamma.hat.all, sigma_sq_all,cor_mat_lists

#### Settings of variables and parameters ####
n1 <- 20 # number of nodes of Gauss-Hermite quadrature for normal distribution
n2 <- 10 # number of nodes of Gauss-Hermite quadrature for bivariate normal distribution
N0 <- nrow(bv) # number of patients
gh <- gauss.hermite(n1) # univariate Gauss-Hermite quadrature nodes and weights
bigh <- bigauss.hermite(n2) # bivariate Gauss-Hermite quadrature nodes and weights
m <- 2 # number of baseline variables
p <- 6 # number of temporal processes(k)

## Canonical links, their inverse functions, and their derivatives for each response(latent temporal process) 
par0.glm.func <- list(gaussian(), poisson(), binomial(),
                      gaussian(), binomial(), binomial()
)

linkfunc.inv <- list(identity, exp, function(z){exp(z)/(1+exp(z))},
                     identity, function(z){exp(z)/(1+exp(z))}, 
                     function(z){exp(z)/(1+exp(z))}
)

linkfunc.inv.deriv <- list(
  function(x){rep(1,length(x))}, 
  exp, 
  function(z){exp(z)/(1+exp(z))^2},
  function(x){rep(1,length(x))}, 
  function(z){exp(z)/(1+exp(z))^2},
  function(z){exp(z)/(1+exp(z))^2}
)

## Specify kernel functions
k_epan <- function(lambda, x, x0){
  t <- abs(x-x0)/lambda
  return(ifelse(t<=1, 3/4*(1-t^2)/lambda, 0))
}
kernelfunc <- list(k_epan, k_epan)

## Specify bandwidths in kernel functions
pow.k <- 3; pow.kl <- 3
bw.k <- 5*N0^(-1/pow.k)
bw.kl <- 10*N0^(-1/pow.kl)
bandwidth.all <- c(bw.k, bw.kl)
bandwidth.all

## Specify variance parameters need to be estimated or not
var.deriv.est <- as.logical(rep("F", p))

## Fix the variances and do not estimate them for now
var.fixed <- sigma_sq_all

## gamma estimators
gamma.hat.all # m*p matrix, each column for each temporal process

#### Settings of objective and gradient function ####
time_Xikj_all_mat <- as.matrix(time_Xikj_all_df)

## time grids
mmm <- 1
ttt <- t.grids[mmm]

#### Beta Estimators ####
par_beta_est <- vector("list",p) # a list of length p, each sublist for each marker
for (kkk in 1:p){
  ## Adjust Y and X(Z) 
  flag <-((time_Xikj_all_mat[,2])==kkk & (time_Xikj_all_mat[,4])<ttt+bandwidth.all[[1]] & (time_Xikj_all_mat[,4])>ttt-bandwidth.all[[1]] )
  ifelse(sum(flag)==1,
         time_Xikj <- t(time_Xikj_all_mat[flag,1:5]),
         time_Xikj <- time_Xikj_all_mat[flag,1:5]
  )
  
  Z_i <- bv %>% dplyr::select(-i) %>% as.matrix
  N <- nrow(Z_i)
  
  ## Initial values of parameters(obtained from glm models)
  closest.X <- closest_Obs(time_Xikj_all_df, ttt, kkk) # NN*1 vector
  
  glmfit <- glm(y~.-1, data=data.frame(y=closest.X, Z_i), family=par0.glm.func[[kkk]])
  par0.beta <- glmfit$coefficients
  names(par0.beta) <- paste0(names(glmfit$coefficients),"_",kkk)
  par.iter <- par0.beta
  
  ## Estimate regression parameters
  tempfile <- file.path(iterdir,paste("reportOptim",casename,"tp",mmm,"mk",kkk,sep="_"))
  sink(tempfile)
  par.beta.est <- optim(par.iter, fn=obj.func.k.v3, gr=grad.func.k.v3, hessian=TRUE,
                        kernel.func.k=kernelfunc[[1]],
                        an.k=bandwidth.all[[1]],
                        t=ttt,
                        k=kkk,
                        Z=Z_i,
                        time_Xikj=time_Xikj,
                        gfun.inv.k=linkfunc.inv[[kkk]],
                        gfun.inv.k.deriv=linkfunc.inv.deriv[[kkk]],
                        gamma.hat.k=gamma.hat.all[,kkk],
                        sigma.sq.k=var.fixed[kkk],
                        var.deriv.k=var.deriv.est[kkk],
                        gh=gh,
                        bigh=bigh,
                        method="L-BFGS-B", control=list(trace=6, maxit=500, factr=1e1))
  sink()
  all_iterOptim <- readLines(tempfile)
  unlink(tempfile)
  
  iterOptim_Parm <- all_iterOptim[grep('^X', all_iterOptim)]

  fname <- paste(casename,"tp",ttt,"marker",kkk,sep="_")
  write(all_iterOptim, file.path(consoledir, paste0(fname,"_alliter.txt")))

  par_beta_est[[kkk]] <- par.beta.est 
}

beta_est <- t(sapply(par_beta_est, function(x){x$par})) # a p*(m+1) matrix, each row is beta for the k-th biomarker


#### Sigma estimators ####
par_cov_est <- vector("list",p*(p-1)/2) # a list of length p, each sublist for each biomarker
for (kkk in 1:(p-1)){
  for (lll in ((kkk+1):p)){  

    ## refine observations 
    time_Xikj_tmp <- time_Xikj_all_df %>% 
      filter(k==kkk & time < ttt+bandwidth.all[[2]] & time > ttt-bandwidth.all[[2]])
    
    time_Xilj_tmp <- time_Xikj_all_df %>% 
      filter(k==lll & time < ttt+bandwidth.all[[2]] & time > ttt-bandwidth.all[[2]])
    
    time_Xikj <- time_Xikj_tmp %>%
      semi_join(time_Xilj_tmp, by = "i") %>%
      as.matrix(rownames=F)
    print(length(unique(time_Xikj[,1])))
    
    time_Xilj <- time_Xilj_tmp %>%
      semi_join(time_Xikj_tmp, by = "i") %>%
      as.matrix(rownames=F)
    print(length(unique(time_Xilj[,1])))
    
    ID_match <- data.frame(ID1=unique(time_Xikj[,1]), ID2=unique(time_Xilj[,1]))
    
    ## refine patients having observations for both markers
    Z_i <- bv %>%
      semi_join(time_Xikj_tmp, by = "i") %>%
      semi_join(time_Xilj_tmp, by = "i")
    print(head(Z_i))
    print(dim(Z_i))
    Z_i %<>% dplyr::select(-i) %>% as.matrix()
    
    N <- nrow(Z_i)
    
    ## Initial values of parameters(obtain from glm models)
    par0.cov <- cor_mat_lists[[mmm]][kkk,lll]*sqrt(var.fixed[kkk]*var.fixed[lll])
    names(par0.cov) <- paste0("sigma_",kkk,"_",lll)
    param.uplmt <- sqrt(var.fixed[kkk]*var.fixed[lll])-1e-3
    param.lowlmt <- -param.uplmt
    
    par.iter <- par0.cov

    ## Estimate correlation parameters
    tempfile <- file.path(iterdir,paste("reportOptim",casename,"tp",mmm,"cov",kkk,lll,sep="_"))
    sink(tempfile)
    par.cov.est <- optim(par.iter, fn=obj.func.kl.v3, gr=grad.func.kl.v3, hessian=TRUE,
                         k=kkk, l=lll,
                         kernel.func.kl=kernelfunc[[1]], an.kl=bandwidth.all[[2]],
                         t=ttt, Z=Z_i, t_X_1=time_Xikj, t_X_2=time_Xilj,
                         beta.k=beta_est[kkk,],
                         gfun.inv.k=linkfunc.inv[[kkk]],
                         gfun.inv.k.deriv=linkfunc.inv.deriv[[kkk]],
                         sigma.sq.k=var.fixed[kkk],
                         gamma.hat.k=gamma.hat.all[,kkk],
                         var.deriv.k=var.deriv.est[kkk],
                         beta.l=beta_est[lll,],
                         gfun.inv.l=linkfunc.inv[[lll]],
                         gfun.inv.l.deriv=linkfunc.inv.deriv[[lll]],
                         sigma.sq.l=var.fixed[lll],
                         gamma.hat.l=gamma.hat.all[,lll],
                         var.deriv.l=var.deriv.est[lll],
                         gh=gh, bigh=bigh,
                         ID_match=ID_match,
                         lower=param.lowlmt, upper=param.uplmt,
                         method="L-BFGS-B", control=list(trace=6, maxit=500, factr=1e1)) # convergence criteria: factr, pgtol
    sink()
    all_iterOptim <- readLines(tempfile)
    unlink(tempfile)
    
    iterOptim_Parm <- all_iterOptim[grep('^X', all_iterOptim)]
    
    fname <- paste(casename,"tp",ttt,"cov",kkk,lll,sep="_")
    write(all_iterOptim, file.path(consoledir, paste0(fname,"_alliter.txt")))

    par_cov_est[[kkk*p-(p-lll)-(1+kkk)*kkk/2]] <- par.cov.est
    
  }}

#### Save all estimators ####
par_est_all <- list(b=par_beta_est, c=par_cov_est)
fname <- paste(casename,"tp",ttt,sep="_")
save(par_est_all, file=file.path(paramdir, paste0(fname,".RData")))

